class RepeatedFlagNameException(Exception):
    def __str__(self):
        return "Repeated registered_flag name in register command"


class TooManyTransferredArgsException(Exception):
    def __str__(self):
        return "Too many transferred arguments"


class RequiredArgumentNotPassedException(Exception):
    def __str__(self):
        return "Required argument not passed"


class IncorrectNumberOfHandlerArgsException(Exception):
    def __str__(self):
        return "Handler has incorrect number of arguments"


class TriggerCannotContainSpacesException(Exception):
    def __str__(self):
        return "Command trigger cannot contain spaces"



from argenta.router import Router


class RegisteredRouters:
    def __init__(self, registered_routers: list[Router] = None) -> None:
        self._registered_routers = registered_routers if registered_routers else []

    def get_registered_routers(self) -> list[Router]:
        return self._registered_routers

    def add_registered_router(self, router: Router):
        self._registered_routers.append(router)

    def add_registered_routers(self, *routers: Router):
        self._registered_routers.extend(routers)

    def __iter__(self):
        return iter(self._registered_routers)

    def __next__(self):
        return next(iter(self._registered_routers))


from typing import Callable, Any
from inspect import getfullargspec

from argenta.command import Command
from argenta.command.models import InputCommand
from argenta.router.command_handlers.entity import CommandHandlers
from argenta.router.command_handler.entity import CommandHandler
from argenta.command.flag.models import Flag, Flags, InputFlags
from argenta.router.exceptions import (RepeatedFlagNameException,
                                     TooManyTransferredArgsException,
                                     RequiredArgumentNotPassedException,
                                     IncorrectNumberOfHandlerArgsException,
                                     TriggerCannotContainSpacesException)


class Router:
    def __init__(self,
                 title: str = 'Commands group title:',
                 name: str = 'Default'):
        self._title = title
        self._name = name

        self._command_handlers: CommandHandlers = CommandHandlers()
        self._ignore_command_register: bool = False

        self._not_valid_flag_handler: Callable[[Flag], None] = lambda flag: print(f"Undefined or incorrect input flag: {flag.get_string_entity()}{(' '+flag.get_value()) if flag.get_value() else ''}")


    def command(self, command: Command) -> Callable[[Any],  Any]:
        self._validate_command(command)

        def command_decorator(func):
            Router._validate_func_args(command, func)
            self._command_handlers.add_command_handler(CommandHandler(func, command))

            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper

        return command_decorator


    def set_invalid_input_flag_handler(self, func):
        processed_args = getfullargspec(func).args
        if len(processed_args) != 1:
            raise IncorrectNumberOfHandlerArgsException()
        else:
            self._not_valid_flag_handler = func


    def input_command_handler(self, input_command: InputCommand):
        input_command_name: str = input_command.get_trigger()
        input_command_flags: InputFlags = input_command.get_input_flags()
        for command_handler in self._command_handlers:
            handle_command = command_handler.get_handled_command()
            if input_command_name.lower() == handle_command.get_trigger().lower():
                if handle_command.get_registered_flags():
                    if input_command_flags:
                        for flag in input_command_flags:
                            is_valid = handle_command.validate_input_flag(flag)
                            if not is_valid:
                                self._not_valid_flag_handler(flag)
                                return
                        command_handler.handling(input_command_flags)
                        return
                    else:
                        command_handler.handling({})
                        return
                else:
                    if input_command_flags:
                        self._not_valid_flag_handler(input_command_flags[0])
                        return
                    else:
                        command_handler.handling()
                        return


    @staticmethod
    def _validate_command(command: Command):
        command_name: str = command.get_trigger()
        if command_name.find(' ') != -1:
            raise TriggerCannotContainSpacesException()

        flags: Flags = command.get_registered_flags()
        if flags:
            flags_name: list = [x.get_string_entity().lower() for x in flags]
            if len(set(flags_name)) < len(flags_name):
                raise RepeatedFlagNameException()


    @staticmethod
    def _validate_func_args(command: Command, func: Callable):
        registered_args = command.get_registered_flags()
        transferred_args = getfullargspec(func).args
        if registered_args.get_flags() and transferred_args:
           if len(transferred_args) != 1:
                raise TooManyTransferredArgsException()
        elif registered_args.get_flags() and not transferred_args:
            raise RequiredArgumentNotPassedException()
        elif not registered_args.get_flags() and transferred_args:
            raise TooManyTransferredArgsException()


    def set_ignore_command_register(self, ignore_command_register: bool):
        self._ignore_command_register = ignore_command_register


    def get_command_handlers(self) -> CommandHandlers:
        return self._command_handlers


    def get_name(self) -> str:
        return self._name


    def get_title(self) -> str:
        return self._title


    def set_title(self, title: str):
        self._title = title


    def get_all_commands(self) -> list[str]:
        all_commands: list[str] = []
        for command_handler in self._command_handlers:
            all_commands.append(command_handler.get_handled_command().get_trigger())

        return all_commands



from argenta.command.flag.models import Flag, InputFlag, Flags, InputFlags
from argenta.command.exceptions import (UnprocessedInputFlagException,
                                         RepeatedInputFlagsException,
                                         EmptyInputCommandException)
from typing import Generic, TypeVar, cast, Literal


InputCommandType = TypeVar('InputCommandType')


class BaseCommand:
    def __init__(self, trigger: str):
        self._trigger = trigger

    def get_trigger(self) -> str:
        return self._trigger


class Command(BaseCommand):
    def __init__(self, trigger: str,
                 description: str = None,
                 flags: Flag | Flags = None):
        super().__init__(trigger)
        self._registered_flags: Flags = flags if isinstance(flags, Flags) else Flags(flags) if isinstance(flags, Flag) else Flags()
        self._description = f'description for "{self._trigger}" command' if not description else description

    def get_registered_flags(self) -> Flags:
        return self._registered_flags

    def validate_input_flag(self, flag: InputFlag):
        registered_flags: Flags | None = self.get_registered_flags()
        if registered_flags:
            if isinstance(registered_flags, Flag):
                if registered_flags.get_string_entity() == flag.get_string_entity():
                    is_valid = registered_flags.validate_input_flag_value(flag.get_value())
                    if is_valid:
                        return True
            else:
                for registered_flag in registered_flags:
                    if registered_flag.get_string_entity() == flag.get_string_entity():
                        is_valid = registered_flag.validate_input_flag_value(flag.get_value())
                        if is_valid:
                            return True
        return False

    def get_description(self) -> str:
        return self._description



class InputCommand(BaseCommand, Generic[InputCommandType]):
    def __init__(self, trigger: str,
                 input_flags: InputFlag | InputFlags = None):
        super().__init__(trigger)
        self._input_flags: InputFlags = input_flags if isinstance(input_flags, InputFlags) else InputFlags(input_flags) if isinstance(input_flags, InputFlag) else InputFlags()

    def _set_input_flags(self, input_flags: InputFlags):
        self._input_flags = input_flags

    def get_input_flags(self) -> InputFlags:
        return self._input_flags

    @staticmethod
    def parse(raw_command: str) -> InputCommandType:
        if not raw_command:
            raise EmptyInputCommandException()

        list_of_tokens = raw_command.split()
        command = list_of_tokens.pop(0)

        input_flags: InputFlags = InputFlags()
        current_flag_name, current_flag_value = None, None

        for k, _ in enumerate(list_of_tokens):
            if _.startswith('-'):
                if current_flag_name or len(_) < 2 or len(_[:_.rfind('-')]) > 3:
                    raise UnprocessedInputFlagException()
                current_flag_name = _
            else:
                if not current_flag_name:
                    raise UnprocessedInputFlagException()
                current_flag_value = _

            if current_flag_name:
                if not len(list_of_tokens) == k+1:
                    if not list_of_tokens[k+1].startswith('-'): continue

                input_flag = InputFlag(name=current_flag_name[current_flag_name.rfind('-')+1:],
                                       prefix=cast(Literal['-', '--', '---'],
                                                   current_flag_name[:current_flag_name.rfind('-')+1]),
                                       value=current_flag_value)

                all_flags = [flag.get_string_entity() for flag in input_flags.get_flags()]
                if input_flag.get_string_entity() not in all_flags:
                    input_flags.add_flag(input_flag)
                else:
                    raise RepeatedInputFlagsException(input_flag)

                current_flag_name, current_flag_value = None, None

        if any([current_flag_name, current_flag_value]):
            raise UnprocessedInputFlagException()
        else:
            return InputCommand(trigger=command, input_flags=input_flags)




from argenta.router.command_handler.entity import CommandHandler


class CommandHandlers:
    def __init__(self, command_handlers: list[CommandHandler] = None):
        self.command_handlers = command_handlers if command_handlers else []

    def get_command_handlers(self) -> list[CommandHandler]:
        return self.command_handlers

    def add_command_handler(self, command_handler: CommandHandler):
        self.command_handlers.append(command_handler)

    def add_command_handlers(self, *command_handlers: CommandHandler):
        self.command_handlers.extend(command_handlers)

    def __iter__(self):
        return iter(self.command_handlers)

    def __next__(self):
        return next(iter(self.command_handlers))


from typing import Callable
from inspect import getfullargspec
import re

from argenta.command.models import Command, InputCommand
from argenta.router import Router
from argenta.router.defaults import system_router
from argenta.command.exceptions import (UnprocessedInputFlagException,
                                  RepeatedInputFlagsException,
                                  EmptyInputCommandException)
from argenta.app.exceptions import (InvalidRouterInstanceException,
                         InvalidDescriptionMessagePatternException,
                         NoRegisteredRoutersException,
                         NoRegisteredHandlersException,
                         IncorrectNumberOfHandlerArgsException)
from argenta.app.registered_routers.entity import RegisteredRouters


class App:
    def __init__(self,
                 prompt: str = 'Enter a command',
                 initial_message: str = '\nHello, I am Argenta\n',
                 farewell_message: str = '\nGoodBye\n',
                 invalid_input_flags_message: str = 'Invalid input flags',
                 exit_command: str = 'Q',
                 exit_command_description: str = 'Exit command',
                 system_points_title: str = 'System points:',
                 ignore_exit_command_register: bool = True,
                 ignore_command_register: bool = False,
                 line_separate: str = '',
                 command_group_description_separate: str = '',
                 repeat_command_groups: bool = True,
                 messages_on_startup: list[str] = None,
                 print_func: Callable[[str], None] = print) -> None:
        self.prompt = prompt
        self.print_func = print_func
        self.exit_command = exit_command
        self.exit_command_description = exit_command_description
        self.system_points_title = system_points_title
        self.ignore_exit_command_register = ignore_exit_command_register
        self.farewell_message = farewell_message
        self.initial_message = initial_message
        self.invalid_input_flags_message = invalid_input_flags_message
        self.line_separate = line_separate
        self.command_group_description_separate = command_group_description_separate
        self.ignore_command_register = ignore_command_register
        self.repeat_command_groups = repeat_command_groups
        self.messages_on_startup = messages_on_startup if messages_on_startup else []

        self._description_message_pattern: str = '[{command}] *=*=* {description}'
        self._registered_routers: RegisteredRouters = RegisteredRouters()
        self._invalid_input_flags_handler: Callable[[str], None] = lambda raw_command: print_func(f'Incorrect flag syntax: "{raw_command}"')
        self._repeated_input_flags_handler: Callable[[str], None] = lambda raw_command: print_func(f'Repeated input flags: "{raw_command}"')
        self._empty_input_command_handler: Callable[[], None] = lambda: print_func('Empty input command')
        self._unknown_command_handler: Callable[[InputCommand], None] = lambda command: print_func(f"Unknown command: {command.get_trigger()}")
        self._exit_command_handler: Callable[[], None] = lambda: print_func(self.farewell_message)


    def start_polling(self) -> None:
        self._setup_system_router()
        self._validate_number_of_routers()
        self._validate_included_routers()

        self.print_func(self.initial_message)

        for message in self.messages_on_startup:
            self.print_func(message)

        if not self.repeat_command_groups:
            self._print_command_group_description()
            self.print_func(self.prompt)

        while True:
            if self.repeat_command_groups:
                self._print_command_group_description()
                self.print_func(self.prompt)

            raw_command: str = input()

            try:
                input_command: InputCommand = InputCommand.parse(raw_command=raw_command)
            except (UnprocessedInputFlagException,
                    RepeatedInputFlagsException,
                    EmptyInputCommandException) as error:
                self.print_func(self.line_separate)
                self._error_handler(error, raw_command)
                self.print_func(self.line_separate)

                if not self.repeat_command_groups:
                    self.print_func(self.prompt)
                continue

            is_exit = self._is_exit_command(input_command)
            if is_exit:
                return

            self.print_func(self.line_separate)
            is_unknown_command: bool = self._check_is_command_unknown(input_command)

            if is_unknown_command:
                self.print_func(self.line_separate)
                self.print_func(self.command_group_description_separate)
                if not self.repeat_command_groups:
                    self.print_func(self.prompt)
                continue

            for registered_router in self._registered_routers:
                registered_router.input_command_handler(input_command)

            self.print_func(self.line_separate)
            self.print_func(self.command_group_description_separate)
            if not self.repeat_command_groups:
                self.print_func(self.prompt)


    def set_initial_message(self, message: str) -> None:
        self.initial_message: str = message


    def set_farewell_message(self, message: str) -> None:
        self.farewell_message: str = message


    def set_description_message_pattern(self, pattern: str) -> None:
        first_check = re.match(r'.*{command}.*', pattern)
        second_check = re.match(r'.*{description}.*', pattern)

        if bool(first_check) and bool(second_check):
            self._description_message_pattern: str = pattern
        else:
            raise InvalidDescriptionMessagePatternException(pattern)


    def set_invalid_input_flags_handler(self, handler: Callable[[str], None]) -> None:
        args = getfullargspec(handler).args
        if len(args) != 1:
            raise IncorrectNumberOfHandlerArgsException()
        else:
            self._invalid_input_flags_handler = handler


    def set_repeated_input_flags_handler(self, handler: Callable[[str], None]) -> None:
        args = getfullargspec(handler).args
        if len(args) != 1:
            raise IncorrectNumberOfHandlerArgsException()
        else:
            self._repeated_input_flags_handler = handler


    def set_unknown_command_handler(self, handler: Callable[[str], None]) -> None:
        args = getfullargspec(handler).args
        if len(args) != 1:
            raise IncorrectNumberOfHandlerArgsException()
        else:
            self._unknown_command_handler = handler


    def set_empty_command_handler(self, handler: Callable[[str], None]) -> None:
        args = getfullargspec(handler).args
        if len(args) != 1:
            raise IncorrectNumberOfHandlerArgsException()
        else:
            self._empty_input_command_handler = handler


    def set_exit_command_handler(self, handler: Callable[[], None]) -> None:
        args = getfullargspec(handler).args
        if len(args) != 0:
            raise IncorrectNumberOfHandlerArgsException()
        else:
            self._exit_command_handler = handler


    def add_message_on_startup(self, message: str) -> None:
        self.messages_on_startup.append(message)


    def include_router(self, router: Router) -> None:
        if not isinstance(router, Router):
            raise InvalidRouterInstanceException()

        router.set_ignore_command_register(self.ignore_command_register)
        self._registered_routers.add_registered_router(router)


    def include_routers(self, *routers: Router) -> None:
        for router in routers:
            self.include_router(router)


    def _validate_number_of_routers(self) -> None:
        if not self._registered_routers:
            raise NoRegisteredRoutersException()


    def _validate_included_routers(self) -> None:
        for router in self._registered_routers:
            if not router.get_command_handlers():
                raise NoRegisteredHandlersException(router.get_name())


    def _setup_system_router(self):
        system_router.set_title(self.system_points_title)
        @system_router.command(Command(self.exit_command, self.exit_command_description))
        def exit_command():
            self._exit_command_handler()
        if system_router not in self._registered_routers.get_registered_routers():
            self.include_router(system_router)


    def _is_exit_command(self, command: InputCommand):
        if command.get_trigger().lower() == self.exit_command.lower():
            if self.ignore_exit_command_register:
                system_router.input_command_handler(command)
                return True
            else:
                if command.get_trigger() == self.exit_command:
                    system_router.input_command_handler(command)
                    return True
        return False


    def _check_is_command_unknown(self, command: InputCommand):
        for router_entity in self._registered_routers:
            for command_handler in router_entity.get_command_handlers():
                handled_command_trigger = command_handler.get_handled_command().get_trigger()
                if handled_command_trigger.lower() == command.get_trigger().lower():
                    if self.ignore_command_register:
                        return False
                    else:
                        if handled_command_trigger == command.get_trigger():
                            return False
        self._unknown_command_handler(command)
        return True


    def _print_command_group_description(self):
        for registered_router in self._registered_routers:
            self.print_func(registered_router.get_title())
            for command_handler in registered_router.get_command_handlers():
                self.print_func(self._description_message_pattern.format(
                        command=command_handler.get_handled_command().get_trigger(),
                        description=command_handler.get_handled_command().get_description()
                    )
                )
            self.print_func(self.command_group_description_separate)


    def _error_handler(self,
                       error: UnprocessedInputFlagException |
                              RepeatedInputFlagsException |
                              EmptyInputCommandException,
                       raw_command: str) -> None:
        match error:
            case UnprocessedInputFlagException():
                self._invalid_input_flags_handler(raw_command)
            case RepeatedInputFlagsException():
                self._repeated_input_flags_handler(raw_command)
            case EmptyInputCommandException():
                self._empty_input_command_handler()





from typing import Literal, Pattern
from abc import ABC, abstractmethod


class BaseFlag:
    def __init__(self, name: str,
                 prefix: Literal['-', '--', '---'] = '--'):
        self._name = name
        self._prefix = prefix

    def get_string_entity(self):
        string_entity: str = self._prefix + self._name
        return string_entity

    def get_name(self):
        return self._name

    def get_prefix(self):
        return self._prefix



class InputFlag(BaseFlag):
    def __init__(self, name: str,
                 prefix: Literal['-', '--', '---'] = '--',
                 value: str = None):
        super().__init__(name, prefix)
        self._flag_value = value

    def get_value(self):
        return self._flag_value

    def set_value(self, value):
        self._flag_value = value



class Flag(BaseFlag):
    def __init__(self, name: str,
                 prefix: Literal['-', '--', '---'] = '--',
                 possible_values: list[str] | Pattern[str] | False = True):
        super().__init__(name, prefix)
        self.possible_values = possible_values

    def validate_input_flag_value(self, input_flag_value: str | None):
        if self.possible_values is False:
            if input_flag_value is None:
                return True
            else:
                return False
        elif isinstance(self.possible_values, Pattern):
            if isinstance(input_flag_value, str):
                is_valid = bool(self.possible_values.match(input_flag_value))
                if bool(is_valid):
                    return True
                else:
                    return False
            else:
                return False

        elif isinstance(self.possible_values, list):
            if input_flag_value in self.possible_values:
                return True
            else:
                return False
        else:
            return True



class BaseFlags(ABC):
    __slots__ = ('_flags',)

    @abstractmethod
    def get_flags(self):
        pass

    @abstractmethod
    def add_flag(self, flag: Flag | InputFlag):
        pass

    @abstractmethod
    def add_flags(self, flags: list[Flag] | list[InputFlag]):
        pass

    @abstractmethod
    def get_flag(self, name: str):
        pass

    def __iter__(self):
        return iter(self._flags)

    def __next__(self):
        return next(iter(self))

    def __getitem__(self, item):
        return self._flags[item]



class Flags(BaseFlags, ABC):
    def __init__(self, *flags: Flag):
        self._flags = flags if flags else []

    def get_flags(self) -> list[Flag]:
        return self._flags

    def add_flag(self, flag: Flag):
        self._flags.append(flag)

    def add_flags(self, flags: list[Flag]):
        self._flags.extend(flags)

    def get_flag(self, name: str) -> Flag | None:
        if name in [flag.get_name() for flag in self._flags]:
            return list(filter(lambda flag: flag.get_name() == name, self._flags))[0]
        else:
            return None



class InputFlags(BaseFlags, ABC):
    def __init__(self, *flags: InputFlag):
        self._flags = flags if flags else []

    def get_flags(self) -> list[InputFlag]:
        return self._flags

    def add_flag(self, flag: InputFlag):
        self._flags.append(flag)

    def add_flags(self, flags: list[InputFlag]):
        self._flags.extend(flags)

    def get_flag(self, name: str) -> InputFlag | None:
        if name in [flag.get_name() for flag in self._flags]:
            return list(filter(lambda flag: flag.get_name() == name, self._flags))[0]
        else:
            return None




class InvalidRouterInstanceException(Exception):
    def __str__(self):
        return "Invalid Router Instance"


class InvalidDescriptionMessagePatternException(Exception):
    def __init__(self, pattern: str):
        self.pattern = pattern
    def __str__(self):
        return ("Invalid Description Message Pattern\n"
                "Correct pattern example: [{command}] *=*=* {description}\n"
                "The pattern must contain two variables: `command` and `description` - description of the command\n"
                f"Your pattern: {self.pattern}")


class NoRegisteredRoutersException(Exception):
    def __str__(self):
        return "No Registered Router Found"


class NoRegisteredHandlersException(Exception):
    def __init__(self, router_name):
        self.router_name = router_name
    def __str__(self):
        return f"No Registered Handlers Found For '{self.router_name}'"


class IncorrectNumberOfHandlerArgsException(Exception):
    def __str__(self):
        return "Incorrect Input Flags Handler has incorrect number of arguments"



from typing import Callable
from argenta.command import Command
from argenta.command.flag.models import InputFlags


class CommandHandler:
    def __init__(self, handler: Callable[[], None] | Callable[[InputFlags], None], handled_command: Command):
        self.handler = handler
        self.handled_command = handled_command

    def handling(self, input_flags: InputFlags):
        if input_flags.get_flags():
            self.handler(input_flags)
        else:
            self.handler()

    def get_handler(self):
        return self.handler

    def get_handled_command(self):
        return self.handled_command


from dataclasses import dataclass
from argenta.command.flag.models import Flag
import re


@dataclass
class PredeterminedFlags:
    HELP = Flag(name='help', possible_values=False)
    SHORT_HELP = Flag(name='h', prefix='-', possible_values=False)

    INFO = Flag(name='info', possible_values=False)
    SHORT_INFO = Flag(name='i', prefix='-', possible_values=False)

    ALL = Flag(name='all', possible_values=False)
    SHORT_ALL = Flag(name='a', prefix='-', possible_values=False)

    HOST = Flag(name='host', possible_values=re.compile(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$'))
    SHORT_HOST = Flag(name='h', prefix='-', possible_values=re.compile(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$'))

    PORT = Flag(name='port', possible_values=re.compile(r'^\d{1,5}$'))
    SHORT_PORT = Flag(name='p', prefix='-', possible_values=re.compile(r'^\d{1,5}$'))
